"""
Interfaces used by the Workspaces' capabilities system.
"""

import inspect
from abc import ABC, abstractmethod
from enum import Enum, auto
from threading import Thread

from .product_interfaces import ProductIF, FutureProductIF
from .schema import CapabilityRequest, Capability, CapabilityExecution

ProductLocator = str
CapabilityName = str


"""

Capability System / Services

"""


class CapabilityInfoIF(ABC):
    """
    Data access object that can look up and record information about capabilities
    and capability requests.
    # FIXME: I opted to make all the methods static. Is that going to work?
    """
    @staticmethod
    @abstractmethod
    def lookup_capability(capability_name: str) -> Capability:
        """
        Look up the definition of a capability
        :param capability_name:  the name of the capability to find
        :return: a capability
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def lookup_capability_request(capability_request_id: int) -> CapabilityRequest:
        """
        Look up a particular request
        :param capability_request_id:  the request identifier
        :return: a capability request
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def save_request(request: CapabilityRequest) -> int:
        """
        Save a capability request and return an integer identifier for it
        :param request:  the request to save
        :return:         the request identifier
        """
        raise NotImplementedError()


class ProjectSettingsIF(ABC):
    """
    Class that deals with projects overriding capability settings and making multiple instances
    of a capability with the same name
    """
    pass


class CapabilityMatrixIF(ABC):
    """
    Maps CASA versions to version-specific templates, allowing us to support a variety of
    CASA versions
    """
    pass


class CapabilityQueueIF(ABC):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """
    @abstractmethod
    def enqueue(self, execution: CapabilityExecution):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def dequeue(self):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def pause(self):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class CapabilityEngineIF(ABC):
    """
    Executes a capability.
    """
    @abstractmethod
    def execute(self, execution: CapabilityExecution):
        pass


class CapabilityServiceIF(ABC):
    """
    The capability service: clients access this to request capability runs
    """
    @abstractmethod
    def run_capability(self, request: CapabilityRequest):
        """
        Run a capability with the requested settings
        :param request: Capability request
        """
        raise NotImplementedError

class EstimationServiceIF(ABC):
    """
    Service that estimates how long a capability will take to complete
    """
    pass


class MetricsServiceIF(ABC):
    """
    Service that records metrics
    """
    pass


class ArchiveServiceIF(ABC):
    """
    Abstracts services that are needed from the archive system.
    """
    @abstractmethod
    def lookup_product(self, locator: ProductLocator) -> "ProductIF":
        """
        Look up a science product by its locator
        :param locator:  science product locator for this product
        :return:         science product
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


FieldName = str
FieldLabel = str


"""

Capability System / Helpers

"""


class QueueRunnerIF(Thread, ABC):
    """
    Manages a capability queue?
    FIXME: Do we need this class? I didn't see reference to it in the diagrams or in the
    FIXME: architecture overview
    """
    pass


class EventTranslatorIF(ABC):
    """
    Translates a Workflow event into a Capability event
    """
    pass


class CapabilityStepIF(ABC):
    """
    A step in a capability sequence
    """
    @abstractmethod
    def execute_against(self, request: CapabilityRequest):
        """
        Execute this capability step
        FIXME: Will capability steps execute themselves? Or will the engine execute them?
        :return: None
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class FutureCapabilityResultIF(ABC, ProductIF):
    """
    Class that represents a product that will in the future result from a completed
    capability execution
    """
    pass


class ParametersIF(ABC):
    """
    Class that represents a set of capability parameters, a required input for the correct
    and complete execution of said capability
    """
    def json(self) -> str:
        """
        Convert to json-formatted string
        :return: JSON string
        """
        raise NotImplementedError


class CapabilityStepType(Enum):
    """
    Enum that specifies the types of CapabilitySteps that are possible
    """
    PrepareAndRunWorkflow = 0
    AwaitQA = auto()
    AwaitWorkflow = auto()
    AwaitProduct = auto()
    AwaitParameter = auto()
    AwaitLargeAllocApproval = auto()
